#!/usr/bin/env python
# -*- coding: koi8-r -*-
"""ALDConverter output lexer generator
"""

import tpg
from ALDC import *

MinorVersion=0.04		# ALDF internal tree format version

class Meta(Parser):
  r"""
token headcode '^(.|\n)*?===+\n' ;
token code '(?=<\n)<\n([^\n]|\n(?!>))+\n>(?=\n)' ;
token comment '#[^\n]*' ;
token tailcode '(?<=\n)===+(.|\n)*' ;
token stringI3 "'''(\n|.)*?'''" ;
token stringII3 '"{3}(\n|.)*?"{3}' ;
token stringII '"([^"]|\\")*"' ;
token stringI "'([^']|\\')*'" ;
separator spaces '\s+' ;

token word '\w+' ;

START/s ->	headcode/h	$ s=[h[:h.rfind('\n=')+1],{}]
	      ( comment*
		Lexem/l		$ s[1].update(l)
		comment*
	      )+
		tailcode/t	$ s.append(t[t.find('\n')+1:])
;
Lexem/l ->	word/t
		comment/c	$ comm=['comment',c[1:].strip()]
		SCode/flags
		SCode/start
		SCode/finish	$ l={t:(comm,flags,start,finish)}
;
SCode/l ->      comment*
	      ( code/c		$ l=['code',self.hackcode(c[2:-2])]
	|	String/l
	      )
	      comment*
;
String/s ->	stringI/c	$ s=['str',c[1:-1]]
	|	stringII/c	$ s=['strc',c[1:-1]]
	|	stringI3/c	$ s=['str3',c[3:-3]]
	|	stringII3/c	$ s=['strc3',c[3:-3]]
;
"""
  verbose=2
  
  def hackcode(self, code):
    """Do some conversions in code:
    ^=		replace with "ret="
    @number	replace with "args[number]"
    @name	replace with "args[2]['name']"
	in all calls:
    @0		is Writer object itself
    @.		is Writer object itself with dot ("args[0].")
    	in case of lexem[2] call:
    @1		is lexem body just parsed
    @2		is mapping of one-level nested body lexems parsed
    @_		is opening string "args[2]['_']", i. e. lexem[1]
    """
    code=tpg.re.sub(r'(^|\n)=',r'\1ret=',code)
    code=code.replace('@.','@0.')
    code=tpg.re.sub(r'@([-\d:]+)(?=[\W])',r'args[\1]',code)
    code=tpg.re.sub(r'@(([\w$])+)',r'args[2]["\1"]',code)
    return code

def printtag(ttype,string):
  """Print tag string as type
  types: comment, code, str, strc, str3, strc3"""
  tagenc={	
  		"comment":	("\t# ",""),
  		"code":		("<\n",	"\n>"),
		"str":		("'",	"'"),
		"strc":		('"',	'"'),
		"str3":		("'''",	"'''"),
		"strc3":	('"""',	'"""') }
  return "%s%s%s"%(tagenc[ttype][0],string,tagenc[ttype][1])

def printlexer(code):
  """Create pylex code for lexer parser
  code = [ python_code, tags, python_code ]
  tags = { "tag": (comment, flags, start, finish), ... }
  """
  ret=["%s======\n"%code[0]]
  lx=[]
  for lex in code[1]:
    lx.append(tuple([printtag(*t) for t in code[1][lex]])+(lex,))
  lx.sort()
  ret+=["%s%s\n%s\n%s\n%s\n"%((l[-1],)+l[:-1]) for l in lx]
  ret.append("\n======\n%s"%code[2])
  return "\n".join(ret)

def printpython(ttype, string, name, stage):
  if ttype != 'code':
    return printtag(ttype, string), None
  lx=('F%s_'%stage)+tpg.re.sub('\W','_',name)
  cd=tpg.re.sub('(^|\n)',r'\1  ',string)+'\n  return ret'
  return lx,cd

def genlexer(code):
  """Create python  code for lexer parser
  code = [ python_code, tags, python_code ]
  tags = { "tag": (comment, flags, start, finish), ... }
  """
  ret=[code[0]]
  lx=["lexer={"]
  functions=[]
  for lex in code[1]:
    comment, flags, start, finish = code[1][lex]
    lx.append('"%s":%s\n('%(lex,printtag(*comment)))
    for el,eln in ((flags, 'f'), (start,'0'), (finish,'1')):
      lexline,funct=printpython(el[0],el[1],lex,eln)
      lx.append("\t%s,"%lexline)
      if funct:
	functions.append("def %s(*args):\n%s\n"%(lexline,funct))
    lx.append("),")
  ret+=functions+lx+["}\n",code[2]]
  return "\n".join(ret)

class Writer:
  def __init__(self, lexer):
    """Create parser from current lexer
    lexer: { lexem: (flags, start, finish), ... }
    """
    self.lexer = lexer
    self.filetypes = {}		# for file types convertion
    self.external = {}		# external objects used
    self.storage = {}		# for names, numbers etc
    self.nbsp = None		# Non-breakable space for '~' flag
    self.prefilters = []	# code filtering functions
    self._pattern = "%s"	# how to print generated text
    self._quote = None		# have no symbols to quote
    self.context=[]		# enclosures context
    self.encoding=None		# will be set when calling

  def prefilter(self, aldst):
    "Filter ALDS tree for some reasons"
    for PreFilter in self.prefilters:
      aldst=PreFilter(aldst)
    return aldst
    
  def __ior__(self, filter):
    "Add filter() to prefilters[]"
    if not callable(filter):
      filter()			# raise an error
    self.prefilters.append(filter)
    return self

  def strcall(self,o,*args):
    "return string(o) or call it"
    if callable(o):
      return str(o(self,*args))
    else:
      return str(o)
  
  def fnconvert(self,fname):
    "Pick writer-compatible type for input file and change the name"
    for ext in [(".".join(fname.split(".")[-i:])).lower() for i in range(fname.count(".")+1,0,-1)]+["*"]:
      if ext in self.filetypes:
	if fname.lower().endswith(self.filetypes[ext]):
	  return fname
	else:
	  return fname+"."+self.filetypes[ext]
    
    raise KeyError, "Cannot find converter for file '%s')"%fname

  def printout(self,alds,divider=" ",wasflags=[""],level=0):
    """Convert alds by lexer, used divider for '>' in wasflags and '<' in flags
    replace divider by self.nbsp if flag exists
    return formatted string and flags
    """
    def findstr(tree):
      "Find nearest string leaf in ALDS tree"
      if len(tree)==1:
	return "***"+str(tree)
      if type(tree[1])==str:
	return tree[1]
      else:
	return findstr(tree[1])

    body=""
    strcall = lambda s: self.strcall(s,body,fmt)
    lexem,fmt=alds[0],{}
    self.context.append(lexem)
    nbsp=self.nbsp or divider
    oldflags,flags=[""],[""]+strcall(self.lexer[lexem][0]).split(' ')
    div='>' in wasflags and '<' in flags and divider or ""
    fmt["_"]=res=strcall(self.lexer[lexem][1])
    #print "#"*(level+1),".".join(wasflags),".".join(flags),lexem,findstr(alds)
    for el in alds[1:]:
      if type(el) is str:
	body+=el
      else:
	fmt[el[0]],oldflags=self.printout(el,divider,oldflags,level+1)
	body+=fmt[el[0]]
    #self.storage.setdefault(lexem,[])
    if 'quote' in flags:	# the only quoting
      body=self.quote(body)
    else:
      for qtype in filter(lambda s: s.startswith("quote="), flags):
	# multiple quoting functions
	# TODO check quote and quote= not to be mixed in the same grammar
	body=self.quote(body,qtype.replace("quote=",""))
    if 'number' in flags:
      self.storage.setdefault(lexem,[]).append(body)
      fmt["$"]=len(self.storage[lexem])
    if 'replace' not in flags: res+=body
    if 'format' in flags:
      res=strcall(self.lexer[lexem][2])
    else:
      res+=strcall(self.lexer[lexem][2])
    if 'filetype' in flags:		# file type conversion is needed
      self.external[res]=self.fnconvert(res)
      res=self.external[res]
    if '>>' in flags:			# pass '>' if any
      # just in case we'll have many '>'- like flags
      flags+=filter(lambda c: c in wasflags, ['>'])
    if 'picktitle' in flags and not self.title:
      self.title=body
    if 'delete' in flags:
      retu='',flags
    else:
      if div:
	if "nbsp" in flags:
	  div=nbsp
	elif "CR" in flags:
	  div="\n"
      retu=div+res,flags
    self.context.pop()
    return retu

  def pattern(self, title, divider):
    """Document to include generated text in"""
    return self._pattern

  def quote(self, string, dst=None):
    """Transform string so that all strange symbols
    will be properly quoted with backslashes"""
    if self._quote:
      if type(self._quote) is regexp:	# old behaviour, dst is "\\\1"
	return quote(string, self._quote, r"\\\1")
      else:
	return quote(string, self._quote, dst)

  def __call__(self,alds,substs=[],title=None,pattern=None,enc=None,divider=" ",**argn):
    """Generic output lexer writer: ALDS tree -> output format specified by lexer
    where flags, start and finish are strings or callable"""
    self.filetypes=substs
    self.title=title
    self.encoding=enc
    if pattern:
      self._pattern=pattern
    text=self.printout(self.prefilter(alds),divider)[0].strip()
    return self.pattern(self.title, divider)%text

  def printobjects(self):
    r=["%s\t<-\t%s"%(self.external[e],e) for e in self.external]
    return "\n".join(r)
      
  formats={}
